#Licensed Materials - Property of IBM
#5724-M24
#Copyright IBM Corporation 2012. All Rights Reserved.
#US Government Users Restricted Rights - Use, duplication or disclosure
#restricted by GSA ADP Schedule Contract with IBM Corp.


import sys
import getopt
from java.util import ResourceBundle
from java.text import MessageFormat;

# keep a handy list of urls for display to the user
deploymentEnvironmentUrls = {}

def getMessage(key, strIfNotFound, args):
	# Allows using internationalized strings, or default embedded
	# if copied to another install
	try:
		bundle = ResourceBundle.getBundle("WLEPIIMessages")
		msg = bundle.getString(key)
		if msg is None:
			return strIfNotFound % args
		else:
			return MessageFormat.format(msg, args)
	except:
		# tried to get the bundle and failed, so return the string
		return strIfNotFound % args
	

def getNodeForServer(server):
	serverName = AdminConfig.showAttribute(server, 'name')
	for node in AdminConfig.getid('/Node:/').splitlines():
		nodeName = AdminConfig.showAttribute(node, 'name')
		servers = AdminConfig.getid('/Node:%s/Server:%s/' % (nodeName, serverName)).splitlines(); 
		if servers:
			return node

def getNodesForCluster(cluster):
	nodes = []
	clusterName = AdminConfig.showAttribute(cluster, 'name')
	members = AdminConfig.getid('/ServerCluster:%s/ClusterMember:/' % clusterName).splitlines()
	for member in members:
		nodeName = AdminConfig.showAttribute(member, 'nodeName')
		node = AdminConfig.getid('/Node:%s/' % nodeName)
		if node not in nodes:
			nodes.append(node)
	return nodes

def getTargetsForApplication(application):
	targets = []
	deployment = AdminConfig.getid('/Deployment:%s/' % application);
	deploymentTargets = AdminUtilities.convertToList(AdminConfig.showAttribute(deployment, 'deploymentTargets'))
	for deploymentTarget in deploymentTargets:
		if AdminConfig.getObjectType(deploymentTarget) == 'ClusteredTarget':
			clusterName = AdminConfig.showAttribute(deploymentTarget, 'name')
			cluster = AdminConfig.getid('/ServerCluster:%s/' % clusterName)
			if cluster not in targets:
				targets.append(cluster)
		elif AdminConfig.getObjectType(deploymentTarget) == 'ServerTarget':
			serverName = AdminConfig.showAttribute(deploymentTarget, 'name')
			nodeName = AdminConfig.showAttribute(deploymentTarget, 'nodeName')
			print getMessage('BPM_PROXY_CFG_APP_DEPLOYED', 'Application %s is deployed to server %s on node %s.', (application, serverName, nodeName))
			server = AdminConfig.getid('/Node:%s/Server:%s/' % (nodeName, serverName))
			if server not in targets:
				targets.append(server)
	return targets


def getNodesWithProxyServer():
	
	# list nodes with proxy servers or cluster members
	results = []
	proxyServers = AdminTask.listServers('[-serverType PROXY_SERVER ]').splitlines()
	for proxyServer in proxyServers:
		node = getNodeForServer(proxyServer)
		if node not in results:
			results.append(node)
	
	return results
	
def getNodesWithoutProxyServer():
	
	# find all nodes
	nodes = AdminConfig.getid('/Node:/').splitlines();
		
	# filter out dmgr node
	nodes.remove(getNodeForServer(AdminTask.listServers('[-serverType DEPLOYMENT_MANAGER ]').splitlines()[0]))
	
	# get list of nodes with a proxy server
	nodesWithProxyServer = getNodesWithProxyServer(); 
	
	# list nodes without proxy servers or cluster members
	results = []
	for node in nodes:
		if node not in nodesWithProxyServer:
			if node not in results:
				results.append(node)
	
	return results

def getProxyServerProfileName(nodeName):
	
	# returns the correct profile name to use for the given node type
	isNodezOS = AdminTask.isNodeZOS(['-nodeName', nodeName]);
	
	if isNodezOS == 'true' :
		return 'proxy_server_foundation_zos'
	else:
		return 'proxy_server_foundation'

def logNodeInformation():
	nodes = AdminConfig.getid('/Node:/').splitlines();
	for node in nodes:
		print getMessage('BPM_PROXY_CFG_FOUND_NODE', 'Found node %s on %s.', (AdminConfig.showAttribute(node, 'name'), AdminConfig.showAttribute(node, 'hostName')))
	
	# log existing proxy server information
	nodesWithProxyServer = getNodesWithProxyServer()
	if nodesWithProxyServer:
		for node in nodesWithProxyServer:
			print getMessage('BPM_PROXY_CFG_FOUND_PROXY', 'Found proxy server on node %s on %s.', (AdminConfig.showAttribute(node, 'name'), AdminConfig.showAttribute(node, 'hostName')))
	else:
		print getMessage('BPM_PROXY_CFG_NO_PROXY_FOUND', 'No existing proxy servers found.', ())
	
	# log list of nodes without a proxy server information
	nodesWithoutProxyServer = getNodesWithoutProxyServer();
	if nodesWithoutProxyServer:
		for node in nodesWithoutProxyServer:
			print getMessage('BPM_PROXY_CFG_NO_PROXY_FOUND_ON_SERVER', 'No proxy servers found on node %s on %s.' ,(AdminConfig.showAttribute(node, 'name'), AdminConfig.showAttribute(node, 'hostName')))
	else:
		print getMessage('BPM_PROXY_CFG_ALL_NODES_HAVE_PROXY', 'All nodes contain a proxy server.', ())
	

def deployBpmProxyServers(proxyNode = None):
	
	topologies = AdminConfig.getid('/WBITopology:/').splitlines()
	configuredTopologies = []
	if topologies:
		for topology in topologies:
			if AdminConfig.showAttribute(topology, 'status') == "Configured":
				configuredTopologies.append(topology)
	else:
		raise getMessage("BPM_PROXY_CFG_NO_DE_FOUND", "No deployment environments found.", ())
	
	if configuredTopologies:
		for topology in configuredTopologies:
			print getMessage("BPM_PROXY_CFG_CONFIGURED_DE_FOUND", "Configured deployment environment found: %s", AdminConfig.showAttribute(topology, 'name'))
	else:
		raise getMessage("BPM_PROXY_CFG_NO_CFG_DE_FOUND", "No configured deployment environments found.", ())

	for topology in configuredTopologies:
		deployBpmProxyServer(topology, proxyNode)
	

def deployBpmProxyServer(topology, proxyNode = None):
	topologyName = AdminConfig.showAttribute(topology, 'name')
	topologyClusters = []
	topologyNodes = []
	
	print getMessage("BPM_PROXY_CFG_FOR_DE", "=== Proxy server for deployment environment '%s' ===", topologyName)
	
	# check if one already exists
	proxyServerName = '%s.Proxy' % topologyName
	if AdminConfig.getid('/Server:%s' % proxyServerName):
		raise getMessage("BPM_PROXY_CFG_ALREADY_EXISTS", "A proxy server named %s already exists.", proxyServerName)
	
	# identify the topology's clusters (in a certain order of preference)
	for role in ["WebApp", "Support", "AppTarget", "Messaging"]:
		wbiCluster = AdminConfig.getid('/WBITopology:%s/WBITopologyRole:%s/WBICluster:/' % (topologyName, role))
		if wbiCluster:
			clusterName = AdminConfig.showAttribute(wbiCluster, 'name')
			cluster = AdminConfig.getid('/ServerCluster:%s' % clusterName);
			if cluster not in topologyClusters:
				topologyClusters.append(cluster)
	
	# identify the topology's nodes 
	for cluster in topologyClusters:
		for node in getNodesForCluster(cluster):
			if node not in topologyNodes:
				topologyNodes.append(node)
	
	# if not already specified, let's pick a proxy node
	if proxyNode == None:
		
		# find a topologyNode that does not already have a proxy server
		for node in topologyNodes:
			if node not in getNodesWithProxyServer():
				proxyNode = node
				break
#		for node in topologyNodes:
#			if node not in nodesWithProxyServer[:]:
#				proxyNode = node
#				nodesWithoutProxyServer.remove(proxyNode)
#				nodesWithProxyServer.append(node)
#				break
		
		# if we haven't found a proxy node yet, pick
		# any node that does not have a proxy server
		if not proxyNode:
			nodesWithoutProxyServer = getNodesWithoutProxyServer()
			if nodesWithoutProxyServer:
				proxyNode = nodesWithoutProxyServer[0]
#		if not proxyNode:
#			if nodesWithoutProxyServer:
#				proxyNode = nodesWithoutProxyServer[0]
#				nodesWithoutProxyServer.remove(proxyNode)
#				nodesWithProxyServer.append(proxyNode)
		
		# if a proxy free node has not yet been found, just pick any
		if not proxyNode:
			if topologyNodes:
				print getMessage("BPM_PROXY_CFG_ALL_NODES_W_PROXY", "All nodes have at least one proxy server.", ())
				proxyNode = topologyNodes[0]
	
	# if somehow we still haven't picked one, get out of here
	if not proxyNode:
		raise getMessage('BPM_PROXY_CFG_COULD_NOT_FIND_NODE', 'Unable to find a suitable node to deploy a proxy server on.', ())
	else:
		print getMessage('BPM_PROXY_CFG_DEPLOY_NODE', 'Proxy Server will be deployed to node %s on %s', (AdminConfig.showAttribute(proxyNode, 'name'), AdminConfig.showAttribute(proxyNode, 'hostName')))
		
	# create the proxy server
	proxyNodeName = AdminConfig.showAttribute(proxyNode, 'name')
	params = {}
	params['-name']= proxyServerName
	params['-templateName']=getProxyServerProfileName(proxyNodeName)
	params['-genUniquePorts'] = 'true'
	params['-selectProtocols']= ['-list', [ 'http', 'sip' ]]
	paramList = []
	[paramList.extend([k,v]) for k,v in params.items()]
	print getMessage("BPM_PROXY_CFG_CREATING_PROXY", "Creating proxy server for the '%s' deployment environment.", topologyName)
	proxy = AdminTask.createProxyServer(proxyNodeName, paramList) 
	# set custom property on proxy server required by business space
	print AdminConfig.getid("/Node:%s/Server:%s/ProxySettings:/" % (proxyNodeName, proxyServerName)) 
	AdminConfig.create('Property', AdminConfig.getid("/Node:%s/Server:%s/ProxySettings:/" % (proxyNodeName, proxyServerName)), [['validationExpression', ""],['name',"cache.query.string"],['description', ""], ['value',"true"],['required',"false"]])
	# get the http port, https port, and host name of the newly proxy server
	proxyHttpPort = None
	proxyHttpsPort = None
	serverEntry = AdminConfig.getid("/ServerEntry:%s/" % proxyServerName)
	namedEndPoints = AdminConfig.list("NamedEndPoint", serverEntry).splitlines()
	for namedEndPoint in namedEndPoints:
		endPoint = AdminConfig.list('EndPoint', namedEndPoint)
		endPointName = AdminConfig.showAttribute(namedEndPoint, 'endPointName')
		if endPointName == "PROXY_HTTP_ADDRESS":
			proxyHttpPort = AdminConfig.showAttribute(endPoint, 'port')
		elif endPointName == "PROXY_HTTPS_ADDRESS":
			proxyHttpsPort = AdminConfig.showAttribute(endPoint, 'port')
	
	proxyHostName = AdminConfig.showAttribute(proxyNode, 'hostName')

	# create a virtual host for the proxy server
	virtualHostName = "%s_host" % proxyServerName
	virtualHost = AdminConfig.create('VirtualHost', AdminConfig.getid('/Cell:/'), [['name', virtualHostName]])
	
	# add host aliases for the proxy server 
	AdminConfig.create('HostAlias', virtualHost, [['hostname', proxyHostName],['port',proxyHttpPort]])
	AdminConfig.create('HostAlias', virtualHost, [['hostname', proxyHostName],['port',proxyHttpsPort]])
	
	# add host aliases for the cluster members
	# this is a work-around for some applications that make server side http requests
	for cluster in topologyClusters:
		for clusterMember in AdminConfig.list('ClusterMember', cluster).splitlines():
			memberName = AdminConfig.showAttribute(clusterMember, "memberName")
			memberNodeName = AdminConfig.showAttribute(clusterMember, "nodeName")
			memberNode = AdminConfig.getid('/Node:%s/' % memberNodeName)
			memberHostName = AdminConfig.showAttribute(memberNode, 'hostName')
			serverEntry = AdminConfig.getid("/Node:%s/ServerIndex:/ServerEntry:%s/" % (memberNodeName, memberName))
			namedEndPoints = AdminConfig.list("NamedEndPoint", serverEntry).splitlines()
			for namedEndPoint in namedEndPoints:
				endPoint = AdminConfig.list('EndPoint', namedEndPoint)
				endPointName = AdminConfig.showAttribute(namedEndPoint, 'endPointName')
				if endPointName == "WC_defaulthost" or endPointName == "WC_defaulthost_secure":
					clusterMemberPort = AdminConfig.showAttribute(endPoint, 'port')
					AdminConfig.create('HostAlias', virtualHost, [['hostname', memberHostName],['port',clusterMemberPort]])
					AdminConfig.create('HostAlias', virtualHost, [['hostname', 'localhost'],['port',clusterMemberPort]])
	
	print getMessage("BPM_PROXY_CFG_CREATING_VH", "Virtual host '%s' created.", virtualHostName)

	# update default_host for the https re-direction limitation workaround
	# setting the hostname to a wildcard ensures that the proxy virtual host is a better match
	default_host = AdminConfig.getid('/VirtualHost:default_host/')
	AdminConfig.create('HostAlias', default_host, [['hostname', '*'],['port',proxyHttpPort]])
	AdminConfig.create('HostAlias', default_host, [['hostname', '*'],['port',proxyHttpsPort]])

	# get list of https ports in use across the topology's cluster members
	clusterMemberHttpsPorts = []
	for cluster in topologyClusters:
		for clusterMember in AdminConfig.list('ClusterMember', cluster).splitlines():
			memberName = AdminConfig.showAttribute(clusterMember, "memberName")
			nodeName = AdminConfig.showAttribute(clusterMember, "nodeName")
			serverEntry = AdminConfig.getid("/Node:%s/ServerIndex:/ServerEntry:%s/" % (nodeName, memberName))
			namedEndPoints = AdminConfig.list("NamedEndPoint", serverEntry).splitlines()
			for namedEndPoint in namedEndPoints:
				endPoint = AdminConfig.list('EndPoint', namedEndPoint)
				endPointName = AdminConfig.showAttribute(namedEndPoint, 'endPointName')
				if endPointName == "WC_defaulthost_secure":
					clusterMemberHttpsPorts.append(AdminConfig.showAttribute(endPoint, 'port'))
					break
	
	# create URL re-writing rules for the https re-direction limitation workaround
	proxySettings = AdminConfig.getid("/Server:%s/ProxySettings:/" % proxyServerName)
	proxyRewritingPolicy = AdminConfig.list("RewritingPolicy", proxySettings)
	if not proxyRewritingPolicy:
		proxyRewritingPolicy = AdminConfig.create("RewritingPolicy",proxySettings, [])
	for httpsPort in clusterMemberHttpsPorts:
		fromURLPattern = "https://%s:%s/*" % (proxyHostName, httpsPort)
		toURLPattern = "https://%s:%s/*" % (proxyHostName, proxyHttpsPort)
		AdminConfig.create('RewritingRule', proxyRewritingPolicy, [["fromURLPattern", fromURLPattern],["toURLPattern", toURLPattern]])
		print getMessage("BPM_PROXY_CFG_CREATING_REWRITE_RULE", "Re-writing rule created to map from '%s' to '%s'.", (fromURLPattern, toURLPattern))

	# map web modules to the virtual host that corresponds to the proxy server
	applications = AdminApp.list().splitlines()
	for application in applications:
		for target in getTargetsForApplication(application):
			if target in topologyClusters:
				print getMessage("BPM_PROXY_CFG_MAPPING_MODULES", "Mapping web modules in application %s to %s", (application, virtualHostName))
				AdminApp.edit(application, ['-MapWebModToVH', [['.*', '.*', virtualHostName]]])
				break;
	
	# compute portal URIs
	urls = {}
	urls['Process Portal'] = "http://%s:%s%s" % (proxyHostName, proxyHttpPort, "/portal")
	urls['Process Portal (Secure)'] = "https://%s:%s%s" % (proxyHostName, proxyHttpsPort, "/portal")
	for item in urls.items():
		print getMessage("BPM_PROXY_CFG_APP_AVAILABLE", "%s is available at %s", item)
		deploymentEnvironmentUrls[topologyName] = urls		


# print table of urls
def printURLs():
	for (topologyName, urls) in deploymentEnvironmentUrls.items():
		print
		print "=" * 60
		print getMessage("BPM_PROXY_CFG_URLS_FOR_DE", " URLs for deployment environment '%s'", topologyName)
		print "=" * 60
		for (application, url) in urls.items():
			print "  %-25s %s" % (application, url)
	
def usage():
	print
	print "Usage:"
	print
	print "   configureProxyServer(.bat|.sh) [OPTIONS]"
	print
	print "   -d, --deployment-environment=DE_NAME Specify the deployment"
	print "                            environment name."
	print "   -n, --node=node_name     Node to deploy proxy server to"
	print "       --no-sync            Do not force the immediate propagation"
	print "                            of changes across nodes."
	print "       --no-save            Do not save configuration changes."
	print "       --help               Display this help"
	print

def main():
	try:
		options, args = getopt.getopt(sys.argv, "d:n:", ["no-save","no-sync", "deployment-environment=", "node=", "help"])
	except getopt.GetoptError, err:
		print str(err)
		usage()
		return 2

	if args:
		usage()
	
	save = 1
	sync = 1
	topologyName = None
	targetNodeName = None
	for option, arg in options:
		if option in ("-d", "--deployment-environment"):
			topologyName = arg
		elif option in ("-n", "--node"):
			targetNodeName = arg
		elif option == "--no-save":
			save = 0
			sync = 0
		elif option == "--no-sync":
			sync = 0
		elif option == "--help":
			usage()
			return
		else:
			assert 0, "un-handled option"
	
	# log existing node information
	logNodeInformation();
	
	# specified node should exist
	if targetNodeName:
		targetNode = AdminConfig.getid('/Node:%s/' % targetNodeName) 
		if not targetNode:
			raise getMessage("BPM_PROXY_CFG_NODE_DOESNOT_EXIST", "Node '%s' does not exist.", targetNodeName)
	else:
		targetNode = None
	
	# deploy proxy server
	if topologyName:
		topology = AdminConfig.getid('/WBITopology:%s/' % topologyName)
		if topology:
			if AdminConfig.showAttribute(topology, 'status') == "Configured":
				deployBpmProxyServer(topology, targetNode)
			else:
				raise getMessage("BPM_PROXY_CFG_DE_NOT_CONFIGURED", "The deployment environment '%s' has not been configured.", topologyName)
		else:
			raise getMessage("BPM_PROXY_CFG_DE_DOESNOT_EXIST", "Deployment environment '%s' does not exist.", topologyName)
	else:
		deployBpmProxyServers(targetNode)
	
	# save and sync
	if save:
		print getMessage("BPM_PROXY_CFG_CONFIG_SAVING", "Configuration is being saved.", ())
		AdminConfig.save()
		if sync:
			print getMessage("BPM_PROXY_CFG_SYNCHRONIZING", "Synchronizing the configuration across all nodes.", ())
			AdminNodeManagement.syncActiveNodes()
		print getMessage("BPM_PROXY_CFG_DONE", "Done.", ())
		
	printURLs()

if __name__ == "__main__":
	sys.exit(main())
